// ** See Copyright Notice at the end of this file

// Provide a bridge from Objective-C types to Lua functions.

#import "LuaBridge.h"
#import "LuaReference.h"
#import "lua.h"
#import "lauxlib.h"
#import <objc/objc-class.h>

SEL selectorForLuaSelector(const char *luaName, BOOL registerIfNotFound)
{
    char *selNameSub;
    int i;
    unsigned selLen = strlen(luaName);
    SEL selector;
    selNameSub = alloca(selLen+1);
    strncpy(selNameSub, luaName, selLen+1);
    selNameSub[selLen] = 0;
    for (i=0;i<selLen;i++) if (selNameSub[i] == '_') selNameSub[i] = ':';
    selector = sel_getUid(selNameSub);
    if ((selector == NULL) && (registerIfNotFound))
    {
        selector = sel_registerName(selNameSub);
    }
    return selector;
}

SEL convertLuaObjectToSEL(LuaState *state, int index)
{
    lua_State *L = [state lua_State];
    const char *name = luaL_check_string(L, index);
    SEL s = selectorForLuaSelector(name, NO);
    if (s == NULL) luaL_argerror(L, index, "not selector");
    return s;
}

Class convertLuaObjectToClass(LuaState *state, int index)
{
    Class idArg;
    lua_State *L = [state lua_State];
    switch (lua_type(L,index))
    {
        int tag;
        case LUA_TSTRING:
            idArg = objc_getClass(luaL_check_string(L, index));
            break;
        case LUA_TUSERDATA:
            tag = lua_tag([state lua_State], index);
            if (tag == state->classTag)
            {
                idArg = *(Class*)lua_touserdata(L, index);
                break;
            }
        case LUA_TNIL:
        case LUA_TNUMBER:
        case LUA_TTABLE:
        case LUA_TFUNCTION:
        default:
            luaL_argerror(L, index, "not class");
            idArg = NULL;
            break;
    }
    return idArg;
}

// This function accepts a Lua object and converts it to Objective-C data of the given type as best it can.
// Once found, it calls the callback. Unfortunately, we can't return just it because the data
// may live in the current stack frame. Yechh!

id convertLuaObjectToID(LuaState *state, int index)
{
    id idArg;
    lua_State *L = [state lua_State];
    switch (lua_type(L,index))
    {
        int tag;
        case LUA_TUSERDATA:
            tag = lua_tag([state lua_State], index);
            if ((tag == state->idTag) || (tag == state->classTag))
            {
                idArg = *(id*)lua_touserdata(L, index);
                break;
            }
            // LAME
            idArg = [LuaReference luaReferenceWithLuaState:state indexToLuaObject:index];
            break;
        case LUA_TNIL:
            idArg = nil;
            break;
        case LUA_TNUMBER:
            idArg = [NSNumber numberWithDouble:lua_tonumber(L, index)];
            break;
        case LUA_TSTRING:
            idArg = [NSString stringWithCString:lua_tostring(L,index) length:lua_strlen(L, index)];
            break;
        case LUA_TTABLE:
        case LUA_TFUNCTION:
            // LAME
            idArg = [LuaReference luaReferenceWithLuaState:state indexToLuaObject:index];
            break;
    }
    return idArg;
}

void convertLuaTypeToObjC(LuaState *state, int luaObjectIndex, const char *argType, CastCallbackType castCallback, void *context1, void *context2)
{
    char charArg;
    short shortArg;
    int intArg;
    long longArg;
    float floatArg;
    double doubleArg;
    id idArg;
    size_t length;
    const char *charPtrArg;
    int returnTypeLength = strlen(argType);
    SEL selArg;
    lua_State *L = [state lua_State];

    switch (argType[returnTypeLength-1])
    {
      case _C_CHR:
      case _C_UCHR:
        charArg = luaL_check_number(L,luaObjectIndex);
        castCallback(&charArg, context1, context2);
        break;
      case _C_SHT:
      case _C_USHT:
        shortArg = luaL_check_number(L,luaObjectIndex);
        castCallback(&shortArg, context1, context2);
        break;
      case _C_INT:
      case _C_UINT:
        intArg = luaL_check_number(L,luaObjectIndex);
        castCallback(&intArg, context1, context2);
        break;
      case _C_LNG:
      case _C_ULNG:
        longArg = luaL_check_number(L,luaObjectIndex);
        castCallback(&longArg, context1, context2);
        break;
      case _C_FLT:
        floatArg = luaL_check_number(L,luaObjectIndex);
        castCallback(&floatArg, context1, context2);
        break;
      case _C_DBL:
        doubleArg = luaL_check_number(L,luaObjectIndex);
        castCallback(&doubleArg, context1, context2);
        break;
      case _C_CHARPTR:
        charPtrArg = luaL_check_lstr(L, luaObjectIndex, &length);
        castCallback(&charPtrArg, context1, context2);
        break;
      case _C_SEL:
        selArg = sel_getUid(luaL_check_string(L, luaObjectIndex));
        castCallback(&selArg, context1, context2);
        break;
      // convert to id
      case _C_ID:
        // several cases depending upon Lua value
        idArg = convertLuaObjectToID(state,luaObjectIndex);
        castCallback(&idArg, context1, context2);
        break;
      case _C_CLASS:
        // several cases depending upon Lua value
        idArg = convertLuaObjectToClass(state,luaObjectIndex);
        castCallback(&idArg, context1, context2);
        break;
/*
#define _C_BFLD         'b'
#define _C_VOID         'v'
#define _C_UNDEF        '?'
#define _C_PTR          '^'
#define _C_ARY_B        '['
#define _C_ARY_E        ']'
#define _C_UNION_B      '('
#define _C_UNION_E      ')'
#define _C_STRUCT_B     '{'
#define _C_STRUCT_E     '}'
*/
  }
}

int convertObjCTypeToLua(void *dataBuffer, const char *objCType, LuaState *state, lua_State *L)
{
    id object, *objectPtr;
    double number;
    const char *selName;
    char *luaName;
    char *s;
    int typeLength = strlen(objCType);
    switch (objCType[typeLength-1])
    {
/*
#define _C_BFLD         'b'
#define _C_UNDEF        '?'
#define _C_PTR          '^'
#define _C_ARY_B        '['
#define _C_ARY_E        ']'
#define _C_UNION_B      '('
#define _C_UNION_E      ')'
#define _C_STRUCT_B     '{'
#define _C_STRUCT_E     '}'
*/
    case _C_SEL:
        selName = sel_getName(*(SEL*)dataBuffer);
        if (selName == NULL) lua_error(L, "Unknown selector");
        s = luaName = (char*)alloca(strlen(selName)+1);
        strcpy(luaName, selName);
        while (*s != 0)
        {
            if (*s == ':') *s = '_';
            s++;
        }
        lua_pushstring(L, luaName);
        return 1;
    case _C_ID:
        object = *(id*)dataBuffer;
        if ([object isMemberOfClass:[LuaReference class]])
        {
            // unwrap Lua object
            [object copyLuaObjectToTopOfStack];
            return 1;
        }
        objectPtr = (id *)lua_newuserdata(L, sizeof(id));
        *objectPtr = object;
        [*objectPtr retain];
        lua_settag(L, state->idTag);
        return 1;
    case _C_CLASS:
        objectPtr = (id *)lua_newuserdata(L, sizeof(id));
        *objectPtr = *(id*)dataBuffer;
        lua_settag(L, state->classTag);
        return 1;
    case _C_UCHR:
        number = *(unsigned char*)dataBuffer;
        goto returnValue;
    case _C_CHR:
        number = *(char*)dataBuffer;
        goto returnValue;
    case _C_SHT:
        number = *(short*)dataBuffer;
        goto returnValue;
    case _C_USHT:
        number = *(unsigned short*)dataBuffer;
        goto returnValue;
    case _C_INT:
        number = *(int*)dataBuffer;
        goto returnValue;
    case _C_UINT:
        number = *(unsigned int*)dataBuffer;
        goto returnValue;
    case _C_LNG:
        number = *(long*)dataBuffer;
        goto returnValue;
    case _C_ULNG:
        number = *(unsigned long*)dataBuffer;
        goto returnValue;
    case _C_FLT:
        number = *(float*)dataBuffer;
        goto returnValue;
    case _C_DBL:
        number = *(double*)dataBuffer;
    returnValue:
        lua_pushnumber(L, number);
        return 1;
    case _C_CHARPTR:
        lua_pushstring(L, *(char**)dataBuffer);
        return 1;
    case _C_VOID:
        return 0;
    }
    return 0;  
}


/*
 * Copyright (c) 2001 By Richard Kiss
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 