// ** See Copyright Notice at the end of this file

// Provide a bridge from Objective-C types to Lua functions.

#import "LuaSubclass.h"

#import <Foundation/Foundation.h>
#import <objc/Protocol.h>

#import "LuaBridge.h"
#import "LuaReference.h"
#import "LuaState.h"

#include "lua.h"
#include "lauxlib.h"

extern id myMsgForward (id self, SEL sel, ...);

static char *luaMethodSignatureForSelectorName(const char *aName);

static NSZone *subclassZone()
{
    static NSZone *sSubclassZone = NULL;
    if (sSubclassZone == NULL) sSubclassZone = NSCreateZone(4096, 4096, NO);
    return sSubclassZone;
}

struct objc_class_with_Lua_Reference
{
    struct objc_class class;
    LuaReference *luaReference;
};

static struct objc_method_list *addMethodToList(struct objc_method_list *r, SEL selector, char *signature, IMP imp)
{
    // realloc. Note that struct objc_method_list includes one struct objc_method
    // so we multiply by the pre-incremented value
    unsigned methodCount = (r == NULL) ? 0 : r->method_count;
    NSZone *zone = subclassZone();
    if (r==NULL) r = (struct objc_method_list*)NSZoneCalloc(zone, 1, sizeof(struct objc_method_list));
    else r = (struct objc_method_list*)NSZoneRealloc(zone, r, sizeof(struct objc_method_list) + sizeof(struct objc_method) * methodCount);
    r->method_list[methodCount].method_name = selector;
    r->method_list[methodCount].method_types = signature;
    r->method_list[methodCount].method_imp = imp;
    r->method_count++;
    return r;
}

LuaReference *luaRefForBridgedObject(id o) 
{
    LuaReference *instance;
    if (o==nil) return nil;
    object_getInstanceVariable(o, "luaReference", (void**)&instance);
    return instance;
}

static id forwardInvocationToLuaIMP(id o, SEL s, NSInvocation *anInvocation)
{
    LuaReference *instance = luaRefForBridgedObject(o);
    LuaState *state;
    lua_State *L;

    state = [instance luaState];
    L = [state lua_State];
    lua_getglobal(L, "forwardInvocation");
    if (lua_isfunction(L,-1))
    {
        convertObjCTypeToLua(&o, @encode(id), state, L);
        convertObjCTypeToLua(&s, @encode(SEL), state, L);
        convertObjCTypeToLua(&anInvocation, @encode(id), state, L);
        lua_call(L, 3, 0);
    }
    return nil;
}

union AllTypes
{
    id *objectPtr;
    double number;
    char charBuffer;
    short shortBuffer;
    int intBuffer;
    long longBuffer;
    float floatBuffer;
    double doubleBuffer;
};

static void setNullImplementationsForFunctionsInTableForClass(Class classClass, LuaReference *luaReference, struct objc_method_list *methodsSoFar)
{
    lua_State *L = [[luaReference luaState] lua_State];
    struct objc_method_list *r = methodsSoFar;
    Class superclass = classClass->super_class;
    IMP dfim = [Object methodFor:@selector(descriptionForInstanceMethod:)];
    [luaReference copyLuaObjectToTopOfStack];
    /* table is in the stack at index -2 */
    lua_pushnil(L);  /* first key */
    while (lua_next(L, -2) != 0)
    {
        SEL theSelector;
        const char *selectorName = lua_tostring(L, -2);
        /* `key' is at index -2 and `value' at index -1 */
        theSelector = selectorForLuaSelector(selectorName, YES);
        if (lua_isfunction(L, -1) && (theSelector != NULL))
        {
            struct objc_method_description *omd = (struct objc_method_description *)dfim(superclass, @selector(descriptionForInstanceMethod:), theSelector);
            char *signature;
            if (omd != NULL) signature = omd->types;
            else signature = luaMethodSignatureForSelectorName(selectorName);
            // add to list
            r = addMethodToList(r, theSelector, signature, myMsgForward);
        }
        lua_pop(L, 1);  /* removes `value'; keeps `index' for next iteration */
    }
    // add imp for forwardInvocation
    r = addMethodToList(r, @selector(forwardInvocation:), "v8@4:8@12", (IMP)forwardInvocationToLuaIMP);
    // any others?

    if (r != NULL) class_addMethods(classClass, r);
}

static id luaClassInit(id o, SEL sel, ...)
{
    struct objc_class_with_Lua_Reference *theClass = (struct objc_class_with_Lua_Reference *)[o class];
    LuaReference *lr = theClass->luaReference, *instance;
    LuaState *state = [lr luaState];
    lua_State *L = [state lua_State];
    [lr copyLuaObjectToTopOfStack];
    lua_getglobal(L, "deepclone");
    lua_pushvalue(L, -2);
    lua_call(L, 1, 1);
    instance = [[LuaReference alloc] initWithLuaState:state indexToLuaObject:-1];
    object_setInstanceVariable(o, "luaReference", instance);

    if ([o respondsToSelector:@selector(didInit)]) objc_msgSend(o, @selector(didInit));
    return o;
}

static id luaClassDealloc(id o, SEL sel, ...)
{
    LuaReference *instance;
    object_getInstanceVariable(o, "luaReference", (void**)&instance);
    [instance release];
    return o;
}

Class subclassLuaBridge(Class superclass, const char *newClassName, LuaReference *classMethodImps)
{
    NSZone *zone = subclassZone();
    struct objc_class_with_Lua_Reference *instanceClassStorage = (struct objc_class_with_Lua_Reference *)NSZoneCalloc(zone, sizeof(struct objc_class_with_Lua_Reference), 1);
    struct objc_class_with_Lua_Reference *classClassStorage = (struct objc_class_with_Lua_Reference *)NSZoneCalloc(zone, sizeof(struct objc_class_with_Lua_Reference), 1);
    Class instanceClass = (Class)instanceClassStorage;
    Class classClass = (Class)classClassStorage;
    Class isa = ((struct objc_class*)objc_getClass("Object"))->isa;
    char *name = (char*)NSZoneMalloc(zone, strlen(newClassName)+1);
    struct objc_ivar_list *luaReferenceIVar;
    struct objc_method_list *methods;

    strcpy(name, newClassName);

    luaReferenceIVar = (struct objc_ivar_list*)NSZoneCalloc(zone, sizeof(struct objc_ivar_list), 1);
    luaReferenceIVar->ivar_count = 1;
    luaReferenceIVar->ivar_list[0].ivar_name = "luaReference";
    luaReferenceIVar->ivar_list[0].ivar_type = "@";
    luaReferenceIVar->ivar_list[0].ivar_offset = superclass->instance_size;

    instanceClass->isa = isa;
    instanceClass->super_class = superclass->isa;
    instanceClass->name = name;
    instanceClass->version = 0;
    instanceClass->instance_size = superclass->instance_size;
    instanceClass->ivars = 0;
    instanceClass->methodLists = (struct objc_method_list**)NSZoneCalloc(zone,sizeof(struct objc_method_list*),1);
    instanceClassStorage->luaReference = [classMethodImps retain];

    classClass->isa = instanceClass;
    classClass->super_class = superclass;
    classClass->name = name;
    classClass->version = 0;
    classClass->instance_size = superclass->instance_size + sizeof(LuaReference*);
    classClass->ivars = luaReferenceIVar;
    classClass->methodLists = (struct objc_method_list**)NSZoneCalloc(zone,sizeof(struct objc_method_list*),1);
    classClassStorage->luaReference = [classMethodImps retain];

    objc_addClass(classClass);

    methods = addMethodToList(NULL, @selector(init), "@4@4:8", luaClassInit);
    methods = addMethodToList(methods, @selector(dealloc), "v4@4:8", luaClassDealloc);

    // iterate over Lua table, look for selectors
    setNullImplementationsForFunctionsInTableForClass(classClass, classMethodImps, methods);

    return classClass;
}

@interface LuaMethodSignature : NSObject{}
@end

@implementation LuaMethodSignature

-(id)selW0Args { return nil; }
-(id)selW1Args:a1 { return nil; }
-(id)selW2Args:a1:a2 { return nil; }
-(id)selW3Args:a1:a2:a3 { return nil; }
-(id)selW4Args:a1:a2:a3:a4 { return nil; }
-(id)selW5Args:a1:a2:a3:a4:a5 { return nil; }
-(id)selW6Args:a1:a2:a3:a4:a5:a6 { return nil; }
-(id)selW7Args:a1:a2:a3:a4:a5:a6:a7 { return nil; }
-(id)selW8Args:a1:a2:a3:a4:a5:a6:a7:a8 { return nil; }
-(id)selW9Args:a1:a2:a3:a4:a5:a6:a7:a8:a9{ return nil; }
-(id)selW10Args:a1:a2:a3:a4:a5:a6:a7:a8:a9:a10 { return nil; }
-(id)selW11Args:a1:a2:a3:a4:a5:a6:a7:a8:a9:a10:a11 { return nil; }
-(id)selW12Args:a1:a2:a3:a4:a5:a6:a7:a8:a9:a10:a11:a12 { return nil; }

static char *luaMethodSignatureForSelectorCount(unsigned count)
{
    IMP dfim = [Object methodFor:@selector(descriptionForInstanceMethod:)];
    struct objc_method_description *omd;
    SEL selector = NULL;
    switch (count)
    {
        case 0: selector = @selector(selW0Args); break;
        case 1: selector = @selector(selW1Args:); break;
        case 2: selector = @selector(selW2Args::); break;
        case 3: selector = @selector(selW3Args:::); break;
        case 4: selector = @selector(selW4Args::::); break;
        case 5: selector = @selector(selW5Args:::::); break;
        case 6: selector = @selector(selW6Args::::::); break;
        case 7: selector = @selector(selW7Args:::::::); break;
        case 8: selector = @selector(selW8Args::::::::); break;
        case 9: selector = @selector(selW9Args:::::::::); break;
        case 10: selector = @selector(selWAArgs::::::::::); break;
        case 11: selector = @selector(selWBArgs:::::::::::); break;
        case 12: selector = @selector(selWCArgs::::::::::::); break;
        default: break;
    }
    omd = (struct objc_method_description *)dfim([LuaMethodSignature class], @selector(descriptionForInstanceMethod:), selector);
    if (omd == NULL) return "";
    return omd->types;
}

static char *luaMethodSignatureForSelectorName(const char *aName)
{
    unsigned n = 0;
    while (*aName != 0)
    {
        if ((*aName == ':') || (*aName == '_')) n++;
        aName++;
    }
    return luaMethodSignatureForSelectorCount(n);
}

@end


/*
 * Copyright (c) 2001 By Richard Kiss
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
 