// ** See Copyright Notice at the end of this file

#import "LuaBridge.h"

#import <Foundation/Foundation.h>

#import <objc/objc-class.h>
#import <objc/objc-runtime.h>
#import <objc/Object.h>
#import <objc/Protocol.h>

#include <stdlib.h>

#include "lua.h"
#include "lauxlib.h"

#import <Foundation/NSObject.h>
#import "LuaReference.h"
#import "LuaState.h"
#import "LuaSubclass.h"

union AllTypes
{
    id *objectPtr;
    double number;
    char charBuffer;
    short shortBuffer;
    int intBuffer;
    long longBuffer;
    float floatBuffer;
    double doubleBuffer;
};

NSString *initLuaFilename = @"init.lua";

@interface NSInvocation(LuaBridge)
-(int)castArgumentAtIndex:(unsigned)index pushOntoLuaState:(LuaState*)state;
-(void)popLuaValueToReturnValue:(LuaState*)aState;
@end

static void setInvocationReturnValue(void *bufferToData, void *invocation, void *unused)
{
    [((NSInvocation*)invocation) setReturnValue:bufferToData];
}

@implementation NSInvocation(LuaBridge)
-(int)castArgumentAtIndex:(unsigned)index pushOntoLuaState:(LuaState*)state;
{
    union AllTypes argValue;
    if (index > [[self methodSignature] numberOfArguments]) return 0;
    [self getArgument:&argValue atIndex:index];
    return convertObjCTypeToLua(&argValue, [[self methodSignature] getArgumentTypeAtIndex:index], state, [state lua_State]);
}

-(void)popLuaValueToReturnValue:(LuaState*)aState
{
    convertLuaTypeToObjC(aState, -1, [[self methodSignature] methodReturnType], setInvocationReturnValue, self, NULL);
}

@end

static void setInvocation(void *bufferPtr, void *invocation, void *index)
{
    [((NSInvocation*)invocation) setArgument:bufferPtr atIndex:*(int*)index];
}

static int sendMessage(lua_State *L)
{
    NSMethodSignature *signature;
    LuaState *state = *(LuaState **)lua_touserdata(L, -1);
    id object = convertLuaObjectToID(state, 1);
    SEL selector = convertLuaObjectToSEL(state, 2);
    NSInvocation *invocation;
    int i,nargs;
    char *buffer;
    const char *objcType;
    unsigned bufferLength;
    const char *exceptionText = NULL;
    
    if (object == NULL) lua_error(L, "can't message non-object");
    if (selector == NULL) luaL_verror(L, "can't find selector %s", selector);
    
    signature = [object methodSignatureForSelector:selector];
    if (signature == NULL) lua_error(L, "can't find selector for object");
    
    nargs = [signature numberOfArguments];
    invocation = [NSInvocation invocationWithMethodSignature:signature];
    
    [invocation setTarget:object];
    [invocation setSelector:selector];
    
    for (i=2;i<nargs;i++) {
        const char *argType = [signature getArgumentTypeAtIndex:i];
        convertLuaTypeToObjC(state, i+1, argType, &setInvocation, invocation, &i);
    }

    NS_DURING
        [invocation invoke];
    NS_HANDLER
        // BRAIN DAMAGE -- how to handle exceptions?
        exceptionText = [[localException description] cString];
    NS_ENDHANDLER
    if (exceptionText != NULL) lua_error(L, exceptionText);

    bufferLength = [signature methodReturnLength];
    buffer = (char*)alloca(bufferLength);
    if (bufferLength>0) [invocation getReturnValue:buffer];
    
    objcType = [signature methodReturnType];

    // BRAIN DAMAGE -- this doesn't work if we've called "alloc"
    // We need to special-case this, and return a lua object tagged "uninited-object" or something
    return convertObjCTypeToLua(buffer, objcType, state, L);
}

static int objcClassForName(lua_State *L)
{
    LuaState *state = *(LuaState **)lua_touserdata(L, -1);
    Class class = convertLuaObjectToClass(state, 1);
    Class *classPtr;
    if (class == nil) return 0;

    classPtr = (Class*)lua_newuserdata(L, sizeof(id));
    *classPtr = class;
    lua_settag(L, state->classTag);
    return 1;
}

static int doNSLogObject(lua_State *L)
{
    LuaState *state = *(LuaState **)lua_touserdata(L, -1);
    NSLog(@"%@", convertLuaObjectToID(state, 1));
    return 0;
}

static int dorelease(lua_State *L)
{
    id o = *(id*)lua_touserdata(L, 1);
    [o release];
    return 0;
}

static int subclass (lua_State *L)
{
    LuaState *luaState = *(LuaState **)lua_touserdata(L, -1);
    LuaReference *lr = [LuaReference luaReferenceWithLuaState:luaState indexToLuaObject:3];
    Class superclass;
    const char *newClassName;
    if (lua_gettop(L) != 4) lua_error(L, "subclass requires 3 arguments");
    if (lua_type(L,3) != LUA_TTABLE) luaL_argerror(L, 3, "Not a table");
    newClassName = luaL_check_string(L, 1);
    if (lua_tag(L,2) != luaState->classTag) luaL_argerror(L, 2, "Not a class");
    superclass = *(Class*)lua_touserdata(L, 2);
    subclassLuaBridge(superclass, newClassName, lr);
    return 0;
}

static int setInvocationReturn(lua_State *L)
{
    LuaState *state = *(LuaState **)lua_touserdata(L, -1);
    id nsi = convertLuaObjectToID(state, 1);
    if (lua_gettop(L) != 3) lua_error(L, "setInvocationReturn requires 2 arguments");
    if (![nsi isMemberOfClass:[NSInvocation class]]) lua_error(L, "first argument must be NSInvocation object");
    lua_pushvalue(L, 2);
    [nsi popLuaValueToReturnValue:state];
    return 0;
}

static int invocationArg(lua_State *L)
{
    LuaState *state = *(LuaState **)lua_touserdata(L, -1);
    int index = luaL_check_number(L, 2);
    id nsi = convertLuaObjectToID(state, 1);
    if (lua_gettop(L) != 3) lua_error(L, "invocationArg requires 2 arguments");
    if (![nsi isMemberOfClass:[NSInvocation class]]) lua_error(L, "first argument must be NSInvocation object");
    return [nsi castArgumentAtIndex:index pushOntoLuaState:state];
}

static int tableForBridgedObject(lua_State *L)
{
    LuaState *state = *(LuaState **)lua_touserdata(L, -1);
    id o = convertLuaObjectToID(state, 1);
    LuaReference *instance = luaRefForBridgedObject(o);
    if (instance == nil) return 0;
    [instance copyLuaObjectToTopOfStack];
    return 1;
}

static const struct luaL_reg gluelibWithClosures[] = {
    {"class", objcClassForName},
    {"message", sendMessage},
    {"subclass", subclass},
    {"po", doNSLogObject},
    // these functions are called from init.lua
    {"invocationArg", invocationArg},
    {"setInvocationReturn", setInvocationReturn},
    {"tableForBridgedObject", tableForBridgedObject},
};


LUALIB_API void opengluelib (lua_State *L, LuaState *ls)
{
  NSString *pathToLuaStartup;
  NSData *luaStartupText;
  LuaState **globals = (LuaState **)lua_newuserdata(L, sizeof(LuaState *));
  unsigned int i;

  *globals = ls;

  for (i=0; i<sizeof(gluelibWithClosures)/sizeof(gluelibWithClosures[0]); i++) {
    /* put `globals' as upvalue for these functions */
    lua_pushvalue(L, -1);
    lua_pushcclosure(L, gluelibWithClosures[i].func, 1);
    lua_setglobal(L, gluelibWithClosures[i].name);
  }

  // override gc for id
  lua_pushvalue(L, -1);
  lua_pushcclosure(L, dorelease, 1);
  lua_settagmethod (L, ls->idTag, "gc");

  lua_pop(L, 1);

  pathToLuaStartup = [[NSBundle bundleForClass:objc_getClass("LuaState")] pathForResource:initLuaFilename ofType:nil];
  luaStartupText = [NSData dataWithContentsOfFile:pathToLuaStartup];
  lua_dobuffer(L, [luaStartupText bytes], [luaStartupText length], [initLuaFilename cString]);
}

/*
 * Copyright (c) 2001 By Richard Kiss
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
 * TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
